@@ -1155,6 +1155,20 @@ static int mcplib_backend_flap_backoff_max(lua_State *L) {
     return 0;
 }
 
+static int mcplib_luagc_ratio(lua_State *L) {
+    float ratio = luaL_checknumber(L, -1);
+    proxy_ctx_t *ctx = PROXY_GET_CTX(L);
+    if (ratio < 1.1) {
+        ratio = 1.1;
+    }
+
+    STAT_L(ctx);
+    ctx->tunables.gc_ratio = ratio;
+    STAT_UL(ctx);
+
+    return 0;
+}
+
 static int mcplib_stat_limit(lua_State *L) {
     proxy_ctx_t *ctx = PROXY_GET_CTX(L);
     int limit = luaL_checkinteger(L, -1);
@@ -1395,13 +1409,13 @@ static int mcplib_log_req(lua_State *L) {
     return 0;
 }
 
-static inline uint32_t _rotl(const uint32_t x, int k) {
+static inline uint32_t _mcp_rotl(const uint32_t x, int k) {
     return (x << k) | (x >> (32 - k));
 }
 
 // xoroshiro128++ 32bit version.
-static uint32_t _nextrand(uint32_t *s) {
-    const uint32_t result = _rotl(s[0] + s[3], 7) + s[0];
+static uint32_t _mcp_nextrand(uint32_t *s) {
+    const uint32_t result = _mcp_rotl(s[0] + s[3], 7) + s[0];
 
     const uint32_t t = s[1] << 9;
 
@@ -1412,7 +1426,7 @@ static uint32_t _nextrand(uint32_t *s) {
 
     s[2] ^= t;
 
-    s[3] = _rotl(s[3], 11);
+    s[3] = _mcp_rotl(s[3], 11);
 
     return result;
 }
@@ -1460,7 +1474,7 @@ static int mcplib_log_reqsample(lua_State *L) {
     } else if (rate > 0) {
         // slightly biased random-to-rate without adding a loop, which is
         // completely fine for this use case.
-        uint32_t rnd = (uint64_t)_nextrand(t->proxy_rng) * (uint64_t)rate >> 32;
+        uint32_t rnd = (uint64_t)_mcp_nextrand(t->proxy_rng) * (uint64_t)rate >> 32;
         if (rnd == 0) {
             do_log = true;
         }
@@ -1673,6 +1687,7 @@ int proxy_register_libs(void *ctx, LIBEVENT_THREAD *t, void *state) {
         {"backend", mcplib_backend},
         {"add_stat", mcplib_add_stat},
         {"ratelim_global_tbf", mcplib_ratelim_global_tbf},
+        {"luagc_ratio", mcplib_luagc_ratio},
         {"stat_limit", mcplib_stat_limit},
         {"backend_connect_timeout", mcplib_backend_connect_timeout},
         {"backend_retry_timeout", mcplib_backend_retry_timeout},
