@@ -2,6 +2,7 @@
 // Functions related to the backend handler thread.
 
 #include "proxy.h"
+#include "proxy_tls.h"
 
 enum proxy_be_failures {
     P_BE_FAIL_TIMEOUT = 0,
@@ -41,12 +42,15 @@ const char *proxy_be_failure_text[] = {
 };
 
 static void proxy_backend_handler(const int fd, const short which, void *arg);
+static void proxy_backend_tls_handler(const int fd, const short which, void *arg);
 static void proxy_beconn_handler(const int fd, const short which, void *arg);
+static void proxy_beconn_tls_handler(const int fd, const short which, void *arg);
 static void proxy_event_handler(evutil_socket_t fd, short which, void *arg);
 static void proxy_event_beconn(evutil_socket_t fd, short which, void *arg);
 static int _prep_pending_write(struct mcp_backendconn_s *be);
 static void _post_pending_write(struct mcp_backendconn_s *be, ssize_t sent);
 static int _flush_pending_write(struct mcp_backendconn_s *be);
+static int _flush_pending_tls_write(struct mcp_backendconn_s *be);
 static void _cleanup_backend(mcp_backend_t *be);
 static void _reset_bad_backend(struct mcp_backendconn_s *be, enum proxy_be_failures err);
 static void _set_main_event(struct mcp_backendconn_s *be, struct event_base *base, int flags, struct timeval *t, event_callback_fn callback);
@@ -91,26 +95,17 @@ static int _proxy_beconn_checkconnect(struct mcp_backendconn_s *be) {
         _reset_bad_backend(be, P_BE_FAIL_CONNECTING);
         return -1;
     }
-    P_DEBUG("%s: backend connected (%s:%s)\n", __func__, be->be_parent->name, be->be_parent->port);
+    P_DEBUG("%s: backend connected [fd: %d] (%s:%s)\n", __func__, mcmc_fd(be->client), be->be_parent->name, be->be_parent->port);
     be->connecting = false;
     be->state = mcp_backend_read;
-    be->bad = false;
+
     // seed the failure time for the flap check.
     gettimeofday(&be->last_failed, NULL);
-    be->depth = 0; // was set to INT_MAX if bad, need to reset.
-    be->failed_count = 0;
 
     be->validating = true;
     // TODO: make validation optional.
 
-    if (_beconn_send_validate(be) == -1) {
-        _reset_bad_backend(be, P_BE_FAIL_BADVALIDATE);
-        return -1;
-    } else {
-        // buffer should be empty during validation stage.
-        assert(be->rbufused == 0);
-        return 0;
-    }
+    return 0;
 }
 
 // Use a simple heuristic to choose a backend connection socket out of a list
@@ -127,21 +122,20 @@ struct mcp_backendconn_s *proxy_choose_beconn(mcp_backend_t *be) {
         // else under low loads only the first conn will ever get used (which
         // is normally good; but sometimes bad if using stateful firewalls)
         for (int x = 0; x < be->conncount; x++) {
-            if (be->be[x].depth == 0) {
-                bec = &be->be[x];
+            struct mcp_backendconn_s *bec_i = &be->be[x];
+            if (bec_i->bad) {
+                continue;
+            }
+            if (bec_i->depth == 0) {
+                bec = bec_i;
                 break;
-            } else if (be->be[x].depth < depth) {
-                depth = bec->depth;
-                bec = &be->be[x];
+            } else if (bec_i->depth < depth) {
+                depth = bec_i->depth;
+                bec = bec_i;
             }
         }
     }
 
-    // drop new requests onto end of conn's io-head, reset the backend one.
-    STAILQ_CONCAT(&bec->io_head, &be->io_head);
-    bec->depth += be->depth;
-    be->depth = 0;
-
     return bec;
 }
 
@@ -167,6 +161,7 @@ static void _proxy_event_handler_dequeue(proxy_event_thread_t *t) {
 
         mcp_backend_t *be = io->backend;
         STAILQ_INSERT_TAIL(&be->io_head, io, io_next);
+        assert(be->depth > -1);
         be->depth++;
         if (!be->stacked) {
             be->stacked = true;
@@ -181,22 +176,28 @@ static void _cleanup_backend(mcp_backend_t *be) {
         // remove any pending events.
         if (!be->tunables.down) {
             int pending = event_pending(&bec->main_event, EV_READ|EV_WRITE|EV_TIMEOUT, NULL);
-            if ((pending & (EV_READ|EV_WRITE|EV_TIMEOUT)) != 0) {
+            if (pending != 0) {
                 event_del(&bec->main_event); // an error to call event_del() without event.
             }
             pending = event_pending(&bec->write_event, EV_READ|EV_WRITE|EV_TIMEOUT, NULL);
-            if ((pending & (EV_READ|EV_WRITE|EV_TIMEOUT)) != 0) {
+            if (pending != 0) {
                 event_del(&bec->write_event); // an error to call event_del() without event.
             }
             pending = event_pending(&bec->timeout_event, EV_TIMEOUT, NULL);
-            if ((pending & (EV_TIMEOUT)) != 0) {
+            if (pending != 0) {
                 event_del(&bec->timeout_event); // an error to call event_del() without event.
             }
 
             // - assert on empty queue
             assert(STAILQ_EMPTY(&bec->io_head));
 
+            mcp_tls_shutdown(bec);
             mcmc_disconnect(bec->client);
+
+            if (bec->bad) {
+                mcp_sharedvm_delta(bec->event_thread->ctx, SHAREDVM_BACKEND_IDX,
+                    bec->be_parent->label, -1);
+            }
         }
         // - free be->client
         free(bec->client);
@@ -214,16 +215,21 @@ static void _setup_backend(mcp_backend_t *be) {
             // backend is "forced" into a bad state. never connect or
             // otherwise attempt to use it.
             be->be[x].bad = true;
-            be->be[x].depth = UINT_MAX / 2;
             continue;
         }
         // assign the initial events to the backend, so we don't have to
         // constantly check if they were initialized yet elsewhere.
         // note these events will not fire until event_add() is called.
         int status = mcmc_connect(bec->client, be->name, be->port, bec->connect_flags);
-        event_assign(&bec->main_event, bec->event_thread->base, mcmc_fd(bec->client), EV_WRITE|EV_TIMEOUT, proxy_beconn_handler, bec);
-        event_assign(&bec->write_event, bec->event_thread->base, mcmc_fd(bec->client), EV_WRITE|EV_TIMEOUT, proxy_backend_handler, bec);
-        event_assign(&bec->timeout_event, bec->event_thread->base, -1, EV_TIMEOUT, proxy_backend_handler, bec);
+        event_callback_fn _beconn_handler = &proxy_beconn_handler;
+        event_callback_fn _backend_handler = &proxy_backend_handler;
+        if (be->tunables.use_tls) {
+            _beconn_handler = &proxy_beconn_tls_handler;
+            _backend_handler = &proxy_backend_tls_handler;
+        }
+        event_assign(&bec->main_event, bec->event_thread->base, mcmc_fd(bec->client), EV_WRITE|EV_TIMEOUT, _beconn_handler, bec);
+        event_assign(&bec->write_event, bec->event_thread->base, mcmc_fd(bec->client), EV_WRITE|EV_TIMEOUT, _backend_handler, bec);
+        event_assign(&bec->timeout_event, bec->event_thread->base, -1, EV_TIMEOUT, _backend_handler, bec);
 
         if (status == MCMC_CONNECTING || status == MCMC_CONNECTED) {
             // if we're already connected for some reason, still push it
@@ -288,6 +294,22 @@ static void proxy_event_beconn(evutil_socket_t fd, short which, void *arg) {
     }
 }
 
+static void _proxy_flush_backend_queue(mcp_backend_t *be) {
+    io_pending_proxy_t *io = NULL;
+    P_DEBUG("%s: fast failing request to bad backend (%s:%s) depth: %d\n", __func__, be->name, be->port, be->depth);
+
+    while (!STAILQ_EMPTY(&be->io_head)) {
+        io = STAILQ_FIRST(&be->io_head);
+        STAILQ_REMOVE_HEAD(&be->io_head, io_next);
+        mcp_resp_set_elapsed(io->client_resp);
+        io->client_resp->status = MCMC_ERR;
+        io->client_resp->resp.code = MCMC_CODE_SERVER_ERROR;
+        be->depth--;
+        assert(be->depth > -1);
+        return_io_pending((io_pending_t *)io);
+    }
+}
+
 void proxy_run_backend_queue(be_head_t *head) {
     mcp_backend_t *be;
     STAILQ_FOREACH(be, head, be_next) {
@@ -295,22 +317,31 @@ void proxy_run_backend_queue(be_head_t *head) {
         int flags = 0;
         struct mcp_backendconn_s *bec = proxy_choose_beconn(be);
 
+        int limit = be->tunables.backend_depth_limit;
         if (bec->bad) {
-            // flush queue if backend is still bad.
-            io_pending_proxy_t *io = NULL;
-            P_DEBUG("%s: fast failing request to bad backend (%s:%s) depth: %d\n", __func__, be->name, be->port, bec->depth);
-            while (!STAILQ_EMPTY(&bec->io_head)) {
-                io = STAILQ_FIRST(&bec->io_head);
-                STAILQ_REMOVE_HEAD(&bec->io_head, io_next);
-                io->client_resp->status = MCMC_ERR;
-                io->client_resp->resp.code = MCMC_CODE_SERVER_ERROR;
-                bec->depth--;
-                return_io_pending((io_pending_t *)io);
-            }
-        } else if (bec->connecting || bec->validating) {
+            // TODO: another counter for fast fails?
+            _proxy_flush_backend_queue(be);
+            continue;
+        } else if (limit && bec->depth > limit) {
+            proxy_ctx_t *ctx = bec->event_thread->ctx;
+            STAT_INCR(ctx, request_failed_depth, be->depth);
+            _proxy_flush_backend_queue(be);
+            continue;
+        }
+
+        // drop new requests onto end of conn's io-head, reset the backend one.
+        STAILQ_CONCAT(&bec->io_head, &be->io_head);
+        bec->depth += be->depth;
+        be->depth = 0;
+
+        if (bec->connecting || bec->validating) {
             P_DEBUG("%s: deferring IO pending connecting (%s:%s)\n", __func__, be->name, be->port);
         } else {
-            flags = _flush_pending_write(bec);
+            if (!bec->ssl) {
+                flags = _flush_pending_write(bec);
+            } else {
+                flags = _flush_pending_tls_write(bec);
+            }
 
             if (flags == -1) {
                 _reset_bad_backend(bec, P_BE_FAIL_WRITING);
@@ -371,27 +402,28 @@ void *proxy_event_thread(void *arg) {
 }
 
 static void _set_main_event(struct mcp_backendconn_s *be, struct event_base *base, int flags, struct timeval *t, event_callback_fn callback) {
-    int pending = event_pending(&be->main_event, EV_READ|EV_WRITE, NULL);
-    if ((pending & (EV_READ|EV_WRITE)) != 0) {
+    int pending = event_pending(&be->main_event, EV_READ|EV_WRITE|EV_TIMEOUT, NULL);
+    if (pending != 0) {
         event_del(&be->main_event); // replace existing event.
     }
 
-    event_assign(&be->main_event, base, mcmc_fd(be->client),
+    int fd = mcmc_fd(be->client);
+    if (fd == 0) {
+        fd = -1; // need to pass -1 to event assign if we're not operating on
+                 // a connection.
+    }
+    event_assign(&be->main_event, base, fd,
             flags, callback, be);
     event_add(&be->main_event, t);
 }
 
 static void _stop_main_event(struct mcp_backendconn_s *be) {
-    int pending = event_pending(&be->main_event, EV_READ|EV_WRITE, NULL);
-    if ((pending & (EV_READ|EV_WRITE|EV_TIMEOUT)) == 0) {
-        return;
-    }
-    event_del(&be->write_event);
+    event_del(&be->main_event);
 }
 
 static void _start_write_event(struct mcp_backendconn_s *be) {
-    int pending = event_pending(&be->main_event, EV_WRITE, NULL);
-    if ((pending & (EV_WRITE|EV_TIMEOUT)) != 0) {
+    int pending = event_pending(&be->write_event, EV_WRITE|EV_TIMEOUT, NULL);
+    if (pending != 0) {
         return;
     }
     // FIXME: wasn't there a write timeout?
@@ -399,10 +431,6 @@ static void _start_write_event(struct mcp_backendconn_s *be) {
 }
 
 static void _stop_write_event(struct mcp_backendconn_s *be) {
-    int pending = event_pending(&be->main_event, EV_WRITE, NULL);
-    if ((pending & (EV_WRITE|EV_TIMEOUT)) == 0) {
-        return;
-    }
     event_del(&be->write_event);
 }
 
@@ -410,7 +438,7 @@ static void _stop_write_event(struct mcp_backendconn_s *be) {
 // persistent listener (optimization + catch disconnects faster)
 static void _start_timeout_event(struct mcp_backendconn_s *be) {
     int pending = event_pending(&be->timeout_event, EV_TIMEOUT, NULL);
-    if ((pending & (EV_TIMEOUT)) != 0) {
+    if (pending != 0) {
         return;
     }
     event_add(&be->timeout_event, &be->tunables.read);
@@ -418,24 +446,22 @@ static void _start_timeout_event(struct mcp_backendconn_s *be) {
 
 static void _stop_timeout_event(struct mcp_backendconn_s *be) {
     int pending = event_pending(&be->timeout_event, EV_TIMEOUT, NULL);
-    if ((pending & (EV_TIMEOUT)) == 0) {
+    if (pending == 0) {
         return;
     }
     event_del(&be->timeout_event);
 }
 
 static void _drive_machine_next(struct mcp_backendconn_s *be, io_pending_proxy_t *p) {
-    struct timeval end;
     // set the head here. when we break the head will be correct.
     STAILQ_REMOVE_HEAD(&be->io_head, io_next);
     be->depth--;
+    assert(p != be->io_next); // don't remove what we need to flush.
+    assert(be->depth > -1);
     be->pending_read--;
+    assert(be->pending_read > -1);
 
-    // stamp the elapsed time into the response object.
-    gettimeofday(&end, NULL);
-    p->client_resp->elapsed = (end.tv_sec - p->client_resp->start.tv_sec) * 1000000 +
-        (end.tv_usec - p->client_resp->start.tv_usec);
-
+    mcp_resp_set_elapsed(p->client_resp);
     // have to do the q->count-- and == 0 and redispatch_conn()
     // stuff here. The moment we call return_io here we
     // don't own *p anymore.
@@ -475,7 +501,7 @@ static int proxy_backend_drive_machine(struct mcp_backendconn_s *be) {
             break;
         case mcp_backend_parse:
             r = p->client_resp;
-            r->status = mcmc_parse_buf(be->client, be->rbuf, be->rbufused, &r->resp);
+            r->status = mcmc_parse_buf(be->rbuf, be->rbufused, &r->resp);
 
             // Quick check if we need more data.
             if (r->resp.code == MCMC_WANT_READ) {
@@ -729,7 +755,11 @@ static void _backend_reconnect(struct mcp_backendconn_s *be) {
     }
     // re-create the write handler for the new file descriptor.
     // the main event will be re-assigned after this call.
-    event_assign(&be->write_event, be->event_thread->base, mcmc_fd(be->client), EV_WRITE|EV_TIMEOUT, proxy_backend_handler, be);
+    event_callback_fn _backend_handler = &proxy_backend_handler;
+    if (be->be_parent->tunables.use_tls) {
+        _backend_handler = &proxy_backend_tls_handler;
+    }
+    event_assign(&be->write_event, be->event_thread->base, mcmc_fd(be->client), EV_WRITE|EV_TIMEOUT, _backend_handler, be);
     // do not need to re-assign the timer event because it's not tied to fd
 }
 
@@ -739,7 +769,11 @@ static void proxy_backend_retry_handler(const int fd, const short which, void *a
     assert(which & EV_TIMEOUT);
     struct timeval tmp_time = be->tunables.connect;
     _backend_reconnect(be);
-    _set_main_event(be, be->event_thread->base, EV_WRITE, &tmp_time, proxy_beconn_handler);
+    event_callback_fn _backend_handler = &proxy_beconn_handler;
+    if (be->be_parent->tunables.use_tls) {
+        _backend_handler = &proxy_beconn_tls_handler;
+    }
+    _set_main_event(be, be->event_thread->base, EV_WRITE, &tmp_time, _backend_handler);
 }
 
 // must be called after _reset_bad_backend(), so the backend is currently
@@ -773,16 +807,21 @@ static void _backend_reschedule(struct mcp_backendconn_s *be) {
         if (!be->bad) {
             P_DEBUG("%s: marking backend as bad\n", __func__);
             STAT_INCR(be->event_thread->ctx, backend_marked_bad, 1);
+            mcp_sharedvm_delta(be->event_thread->ctx, SHAREDVM_BACKEND_IDX,
+                    be->be_parent->label, 1);
             LOGGER_LOG(NULL, LOG_PROXYEVENTS, LOGGER_PROXY_BE_ERROR, NULL, badtext, be->be_parent->name, be->be_parent->port, be->be_parent->label, 0, NULL, 0, retry_time);
         }
         be->bad = true;
-        be->depth = INT_MAX/2; // fast-path cache for "bad" marker
        _set_main_event(be, be->event_thread->base, EV_TIMEOUT, &tmp_time, proxy_backend_retry_handler);
     } else {
         struct timeval tmp_time = be->tunables.connect;
         STAT_INCR(be->event_thread->ctx, backend_failed, 1);
         _backend_reconnect(be);
-        _set_main_event(be, be->event_thread->base, EV_WRITE, &tmp_time, proxy_beconn_handler);
+        event_callback_fn _backend_handler = &proxy_beconn_handler;
+        if (be->be_parent->tunables.use_tls) {
+            _backend_handler = &proxy_beconn_tls_handler;
+        }
+        _set_main_event(be, be->event_thread->base, EV_WRITE, &tmp_time, _backend_handler);
     }
 }
 
@@ -836,8 +875,8 @@ static void _backend_flap_check(struct mcp_backendconn_s *be, enum proxy_be_fail
 // _must_ be called from within the event thread.
 static void _reset_bad_backend(struct mcp_backendconn_s *be, enum proxy_be_failures err) {
     io_pending_proxy_t *io = NULL;
-    P_DEBUG("%s: resetting bad backend: %s\n", __func__, proxy_be_failure_text[err]);
-    // Can't use STAILQ_FOREACH() since return_io_pending() free's the current
+    P_DEBUG("%s: resetting bad backend: [fd: %d] %s\n", __func__, mcmc_fd(be->client), proxy_be_failure_text[err]);
+    // Can't use STAILQ_FOREACH() since r_io_p() free's the current
     // io. STAILQ_FOREACH_SAFE maybe?
     int depth = be->depth;
     while (!STAILQ_EMPTY(&be->io_head)) {
@@ -845,9 +884,11 @@ static void _reset_bad_backend(struct mcp_backendconn_s *be, enum proxy_be_failu
         STAILQ_REMOVE_HEAD(&be->io_head, io_next);
         // TODO (v2): Unsure if this is the best way of surfacing errors to lua,
         // but will do for V1.
+        mcp_resp_set_elapsed(io->client_resp);
         io->client_resp->status = MCMC_ERR;
         io->client_resp->resp.code = MCMC_CODE_SERVER_ERROR;
         be->depth--;
+        assert(be->depth > -1);
         return_io_pending((io_pending_t *)io);
     }
 
@@ -866,6 +907,7 @@ static void _reset_bad_backend(struct mcp_backendconn_s *be, enum proxy_be_failu
     _stop_write_event(be);
     _stop_main_event(be);
     _stop_timeout_event(be);
+    mcp_tls_shutdown(be);
     mcmc_disconnect(be->client);
     // we leave the main event alone, because be_failed() always overwrites.
 
@@ -945,8 +987,18 @@ static void _post_pending_write(struct mcp_backendconn_s *be, ssize_t sent) {
     } // for
 
     // resume the flush from this point.
-    if (io != NULL && !io->flushed) {
-        be->io_next = io;
+    if (io != NULL) {
+        if (!io->flushed) {
+            be->io_next = io;
+        } else {
+            // Check for incomplete list because we hit the iovcnt limit.
+            io_pending_proxy_t *nio = STAILQ_NEXT(io, io_next);
+            if (nio != NULL && !nio->flushed) {
+                be->io_next = nio;
+            } else {
+                be->io_next = NULL;
+            }
+        }
     } else {
         be->io_next = NULL;
     }
@@ -981,15 +1033,135 @@ static int _flush_pending_write(struct mcp_backendconn_s *be) {
     return flags;
 }
 
-// Libevent handler for backends in a connecting state.
-static void proxy_beconn_handler(const int fd, const short which, void *arg) {
+static int _flush_pending_tls_write(struct mcp_backendconn_s *be) {
+    int flags = 0;
+    // Allow us to be called with an empty stack to prevent dev errors.
+    if (STAILQ_EMPTY(&be->io_head)) {
+        return 0;
+    }
+
+    int iovcnt = _prep_pending_write(be);
+
+    int sent = mcp_tls_writev(be, iovcnt);
+    if (sent > 0) {
+        _post_pending_write(be, sent);
+        // FIXME: can _post_pending_write do this and return EV_WRITE?
+        // still have unflushed pending IO's, check for write and re-loop.
+        if (be->io_next) {
+            be->can_write = false;
+            flags |= EV_WRITE;
+        }
+    } else if (sent == MCP_TLS_NEEDIO) {
+        // want io
+        be->can_write = false;
+        flags |= EV_WRITE;
+    } else if (sent == MCP_TLS_ERR) {
+        // hard error from tls
+        flags = -1;
+    }
+
+    return flags;
+}
+
+
+static void proxy_bevalidate_tls_handler(const int fd, const short which, void *arg) {
     assert(arg != NULL);
     struct mcp_backendconn_s *be = arg;
     int flags = EV_TIMEOUT;
     struct timeval tmp_time = be->tunables.read;
 
     if (which & EV_TIMEOUT) {
-        P_DEBUG("%s: backend timed out while connecting\n", __func__);
+        P_DEBUG("%s: backend timed out while connecting [fd: %d]\n", __func__, mcmc_fd(be->client));
+        if (be->connecting) {
+            _reset_bad_backend(be, P_BE_FAIL_CONNTIMEOUT);
+        } else {
+            _reset_bad_backend(be, P_BE_FAIL_READVALIDATE);
+        }
+        return;
+    }
+
+    if (which & EV_READ) {
+        int read = mcp_tls_read(be);
+
+        if (read > 0) {
+            mcmc_resp_t r;
+
+            int status = mcmc_parse_buf(be->rbuf, be->rbufused, &r);
+            if (status == MCMC_ERR) {
+                // Needed more data for a version line, somehow. I feel like
+                // this should set off some alarms, but it is possible.
+                if (r.code == MCMC_WANT_READ) {
+                    _set_main_event(be, be->event_thread->base, EV_READ, &tmp_time, proxy_bevalidate_tls_handler);
+                    return;
+                }
+
+                _reset_bad_backend(be, P_BE_FAIL_READVALIDATE);
+                return;
+            }
+
+            if (r.code != MCMC_CODE_VERSION) {
+                _reset_bad_backend(be, P_BE_FAIL_BADVALIDATE);
+                return;
+            }
+
+            be->validating = false;
+            be->rbufused = 0;
+        } else if (read == 0) {
+            // not connected or error.
+            _reset_bad_backend(be, P_BE_FAIL_DISCONNECTED);
+            return;
+        } else if (read == MCP_TLS_NEEDIO) {
+            // try again failure.
+            _set_main_event(be, be->event_thread->base, EV_READ, &tmp_time, proxy_bevalidate_tls_handler);
+            return;
+        } else if (read == MCP_TLS_ERR) {
+            // hard failure.
+            _reset_bad_backend(be, P_BE_FAIL_READING);
+            return;
+        }
+
+        // Passed validation, don't need to re-read, flush any pending writes.
+        int res = _flush_pending_tls_write(be);
+        if (res == -1) {
+            _reset_bad_backend(be, P_BE_FAIL_WRITING);
+            return;
+        }
+        if (flags & EV_WRITE) {
+            _start_write_event(be);
+        }
+        if (be->pending_read) {
+            _start_timeout_event(be);
+        }
+    }
+
+    // switch to the primary persistent read event.
+    if (!be->validating) {
+        _set_main_event(be, be->event_thread->base, EV_READ|EV_PERSIST, NULL, proxy_backend_tls_handler);
+
+        // we're happily validated and switching to normal processing, so
+        // _now_ the backend is no longer "bad".
+        // If we reset the failed count earlier we then can fail the
+        // validation loop indefinitely without ever being marked bad.
+        if (be->bad) {
+            // was bad, need to mark as no longer bad in shared space.
+            mcp_sharedvm_delta(be->event_thread->ctx, SHAREDVM_BACKEND_IDX,
+                    be->be_parent->label, -1);
+        }
+        be->bad = false;
+        be->failed_count = 0;
+    }
+}
+
+// Libevent handler when we're in TLS mode. Unfortunately the code is
+// different enough to warrant its own function.
+static void proxy_beconn_tls_handler(const int fd, const short which, void *arg) {
+    assert(arg != NULL);
+    struct mcp_backendconn_s *be = arg;
+    //int flags = EV_TIMEOUT;
+    struct timeval tmp_time = be->tunables.read;
+
+    if (which & EV_TIMEOUT) {
+        P_DEBUG("%s: backend timed out while connecting [fd: %d]\n", __func__, mcmc_fd(be->client));
         if (be->connecting) {
             _reset_bad_backend(be, P_BE_FAIL_CONNTIMEOUT);
         } else {
@@ -1005,6 +1177,62 @@ static void proxy_beconn_handler(const int fd, const short which, void *arg) {
             if (_proxy_beconn_checkconnect(be) == -1) {
                 return;
             }
+            // TODO: check return code.
+            mcp_tls_connect(be);
+            // fall through to handshake attempt.
+        }
+    }
+
+    assert(be->validating);
+    int ret = mcp_tls_handshake(be);
+    if (ret == MCP_TLS_NEEDIO) {
+        // Need to try again.
+        _set_main_event(be, be->event_thread->base, EV_READ, &tmp_time, proxy_beconn_tls_handler);
+        return;
+    } else if (ret == 1) {
+        // handshake complete.
+        if (mcp_tls_send_validate(be) != MCP_TLS_OK) {
+            _reset_bad_backend(be, P_BE_FAIL_BADVALIDATE);
+            return;
+        }
+
+        // switch to another handler for the final stage.
+        _set_main_event(be, be->event_thread->base, EV_READ, &tmp_time, proxy_bevalidate_tls_handler);
+    } else if (ret < 0) {
+        // FIXME: FAIL_HANDSHAKE
+        _reset_bad_backend(be, P_BE_FAIL_BADVALIDATE);
+        return;
+    }
+}
+
+// Libevent handler for backends in a connecting state.
+static void proxy_beconn_handler(const int fd, const short which, void *arg) {
+    assert(arg != NULL);
+    struct mcp_backendconn_s *be = arg;
+    int flags = EV_TIMEOUT;
+    struct timeval tmp_time = be->tunables.read;
+
+    if (which & EV_TIMEOUT) {
+        P_DEBUG("%s: backend timed out while connecting [fd: %d]\n", __func__, mcmc_fd(be->client));
+        if (be->connecting) {
+            _reset_bad_backend(be, P_BE_FAIL_CONNTIMEOUT);
+        } else {
+            _reset_bad_backend(be, P_BE_FAIL_READVALIDATE);
+        }
+        return;
+    }
+
+    if (which & EV_WRITE) {
+        be->can_write = true;
+
+        if (be->connecting) {
+            if (_proxy_beconn_checkconnect(be) == -1) {
+                return;
+            }
+            if (_beconn_send_validate(be) == -1) {
+                _reset_bad_backend(be, P_BE_FAIL_BADVALIDATE);
+                return;
+            }
             _set_main_event(be, be->event_thread->base, EV_READ, &tmp_time, proxy_beconn_handler);
         }
 
@@ -1028,7 +1256,7 @@ static void proxy_beconn_handler(const int fd, const short which, void *arg) {
             mcmc_resp_t r;
             be->rbufused += read;
 
-            int status = mcmc_parse_buf(be->client, be->rbuf, be->rbufused, &r);
+            int status = mcmc_parse_buf(be->rbuf, be->rbufused, &r);
             if (status == MCMC_ERR) {
                 // Needed more data for a version line, somehow. I feel like
                 // this should set off some alarms, but it is possible.
@@ -1068,7 +1296,7 @@ static void proxy_beconn_handler(const int fd, const short which, void *arg) {
             _reset_bad_backend(be, P_BE_FAIL_WRITING);
             return;
         }
-        if (flags & EV_WRITE) {
+        if (res & EV_WRITE) {
             _start_write_event(be);
         }
         if (be->pending_read) {
@@ -1079,6 +1307,75 @@ static void proxy_beconn_handler(const int fd, const short which, void *arg) {
     // switch to the primary persistent read event.
     if (!be->validating) {
         _set_main_event(be, be->event_thread->base, EV_READ|EV_PERSIST, NULL, proxy_backend_handler);
+
+        // we're happily validated and switching to normal processing, so
+        // _now_ the backend is no longer "bad".
+        // If we reset the failed count earlier we then can fail the
+        // validation loop indefinitely without ever being marked bad.
+        if (be->bad) {
+            // was bad, need to mark as no longer bad in shared space.
+            mcp_sharedvm_delta(be->event_thread->ctx, SHAREDVM_BACKEND_IDX,
+                    be->be_parent->label, -1);
+        }
+        be->bad = false;
+        be->failed_count = 0;
+    }
+}
+
+static void proxy_backend_tls_handler(const int fd, const short which, void *arg) {
+    struct mcp_backendconn_s *be = arg;
+
+    if (which & EV_TIMEOUT) {
+        P_DEBUG("%s: timeout received, killing backend queue\n", __func__);
+        _reset_bad_backend(be, P_BE_FAIL_TIMEOUT);
+        return;
+    }
+
+    if (which & EV_WRITE) {
+        be->can_write = true;
+        int res = _flush_pending_tls_write(be);
+        if (res == -1) {
+            _reset_bad_backend(be, P_BE_FAIL_WRITING);
+            return;
+        }
+        if (res & EV_WRITE) {
+            _start_write_event(be);
+        }
+    }
+
+    if (which & EV_READ) {
+        // got a read event, always kill the pending read timer.
+        _stop_timeout_event(be);
+        // We do the syscall here before diving into the state machine to allow a
+        // common code path for io_uring/epoll/tls/etc
+        int read = mcp_tls_read(be);
+        if (read > 0) {
+            int res = proxy_backend_drive_machine(be);
+            if (res != 0) {
+                _reset_bad_backend(be, res);
+                return;
+            }
+        } else if (read == 0) {
+            // not connected or error.
+            _reset_bad_backend(be, P_BE_FAIL_DISCONNECTED);
+            return;
+        } else if (read == MCP_TLS_NEEDIO) {
+            // sit on epoll again.
+            return;
+        } else if (read == MCP_TLS_ERR) {
+            _reset_bad_backend(be, P_BE_FAIL_READING);
+            return;
+        }
+
+#ifdef PROXY_DEBUG
+        if (!STAILQ_EMPTY(&be->io_head)) {
+            P_DEBUG("backend has leftover IOs: %d\n", be->depth);
+        }
+#endif
+    }
+
+    if (be->pending_read) {
+        _start_timeout_event(be);
     }
 }
 
