@@ -14,21 +14,17 @@
 #endif
 #include <assert.h>
 #include <stdio.h>
-#include <errno.h>
 #include <stdlib.h>
 #include <string.h>
 #include <pthread.h>
 
 #include "queue.h"
+#include "tls.h"
 
 #ifdef __sun
 #include <atomic.h>
 #endif
 
-#ifdef TLS
-#include <openssl/ssl.h>
-#endif
-
 #define ITEMS_PER_ALLOC 64
 
 /* An item in the connection queue. */
@@ -161,7 +157,7 @@ void pause_threads(enum pause_thread_types type) {
 
     switch (type) {
         case PAUSE_ALL_THREADS:
-            slabs_rebalancer_pause();
+            slab_maintenance_pause(settings.slab_rebal);
             lru_maintainer_pause();
             lru_crawler_pause();
 #ifdef EXTSTORE
@@ -173,7 +169,7 @@ void pause_threads(enum pause_thread_types type) {
             pthread_mutex_lock(&worker_hang_lock);
             break;
         case RESUME_ALL_THREADS:
-            slabs_rebalancer_resume();
+            slab_maintenance_resume(settings.slab_rebal);
             lru_maintainer_resume();
             lru_crawler_resume();
 #ifdef EXTSTORE
@@ -243,7 +239,7 @@ void stop_threads(void) {
             fprintf(stderr, "stopped maintainer\n");
     }
     if (settings.slab_reassign) {
-        stop_slab_maintenance_thread();
+        stop_slab_maintenance_thread(settings.slab_rebal);
         if (settings.verbose > 0)
             fprintf(stderr, "stopped slab mover\n");
     }
@@ -633,12 +629,10 @@ static void thread_libevent_process(evutil_socket_t fd, short which, void *arg)
                             fprintf(stderr, "Can't listen for events on fd %d\n",
                                 item->sfd);
                         }
-#ifdef TLS
                         if (item->ssl) {
-                            SSL_shutdown(item->ssl);
-                            SSL_free(item->ssl);
+                            ssl_conn_close(item->ssl);
+                            item->ssl = NULL;
                         }
-#endif
                         close(item->sfd);
                     }
                 } else {
